@@ -19,8 +19,10 @@
 #ifndef	GRUB_PCI_H
 #define	GRUB_PCI_H	1
 
+#ifndef ASM_FILE
 #include <grub/types.h>
 #include <grub/symbol.h>
+#endif
 
 #define  GRUB_PCI_ADDR_SPACE_MASK	0x01
 #define  GRUB_PCI_ADDR_SPACE_MEMORY	0x00
@@ -66,9 +68,41 @@
 #define  GRUB_PCI_REG_MIN_GNT      0x3e
 #define  GRUB_PCI_REG_MAX_LAT      0x3f
 
+#define  GRUB_PCI_COMMAND_IO_ENABLED    0x0001
+#define  GRUB_PCI_COMMAND_MEM_ENABLED   0x0002
+#define  GRUB_PCI_COMMAND_BUS_MASTER    0x0004
+#define  GRUB_PCI_COMMAND_PARITY_ERROR  0x0040
+#define  GRUB_PCI_COMMAND_SERR_ENABLE   0x0100
+
+#define  GRUB_PCI_STATUS_CAPABILITIES      0x0010
+#define  GRUB_PCI_STATUS_66MHZ_CAPABLE     0x0020
+#define  GRUB_PCI_STATUS_FAST_B2B_CAPABLE  0x0080
+
+#define  GRUB_PCI_STATUS_DEVSEL_TIMING_SHIFT 9
+#define  GRUB_PCI_STATUS_DEVSEL_TIMING_MASK 0x0600
+#define  GRUB_PCI_CLASS_SUBCLASS_VGA  0x0300
+#define  GRUB_PCI_CLASS_SUBCLASS_USB  0x0c03
+
+#ifndef ASM_FILE
+
+enum
+  {
+    GRUB_PCI_CLASS_NETWORK = 0x02
+  };
+enum
+  {
+    GRUB_PCI_CAP_POWER_MANAGEMENT = 0x01
+  };
+
+enum
+  {
+    GRUB_PCI_VENDOR_BROADCOM = 0x14e4
+  };
+
+
 typedef grub_uint32_t grub_pci_id_t;
 
-#ifdef GRUB_UTIL
+#ifdef GRUB_MACHINE_EMU
 #include <grub/pciutils.h>
 #else
 typedef grub_uint32_t grub_pci_address_t;
@@ -99,12 +133,20 @@ grub_pci_get_function (grub_pci_device_t dev)
 #include <grub/cpu/pci.h>
 #endif
 
-typedef int NESTED_FUNC_ATTR (*grub_pci_iteratefunc_t)
-     (grub_pci_device_t dev, grub_pci_id_t pciid);
+typedef int (*grub_pci_iteratefunc_t)
+     (grub_pci_device_t dev, grub_pci_id_t pciid, void *data);
 
 grub_pci_address_t EXPORT_FUNC(grub_pci_make_address) (grub_pci_device_t dev,
 						       int reg);
 
-void EXPORT_FUNC(grub_pci_iterate) (grub_pci_iteratefunc_t hook);
+void EXPORT_FUNC(grub_pci_iterate) (grub_pci_iteratefunc_t hook,
+				    void *hook_data);
+
+#include <grub/dma.h>
+
+grub_uint8_t
+EXPORT_FUNC (grub_pci_find_capability) (grub_pci_device_t dev, grub_uint8_t cap);
+
+#endif
 
 #endif /* GRUB_PCI_H */
